import json
import logging
from bson import ObjectId
from urllib.parse import urljoin, urlparse
import requests

from bs4 import BeautifulSoup

logger = logging.getLogger("uvicorn.error")

class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        return json.JSONEncoder.default(self, o)

def retrieve_system_message(messages):
    """Retrieve the system message"""
    try:
        filtered_messages = list(filter(lambda message: message['role'] == 'system', messages))
        return filtered_messages[0]['content']
    except IndexError:
        return None

def retrieve_chat_messages(messages):
    """Retrieve the chat messages"""
    return [
        (msg["content"]) for msg in messages if msg["role"] in ["user", "assistant"]
    ]

def match_strings(keys: list[str], functions):
    """Match the strings in the keys to the functions"""
    # Initialize array to store output
    output = []

    # Loop through the functions array
    for function in functions:
        # If name property of function matches one of the strings in keys
        if function['name'] in keys:
            # Append the function to the output array
            output.append(function)

    # Return the output array
    return output

def get_links(url: str):
	response = requests.get(url, timeout=5)
	soup = BeautifulSoup(response.text, 'html.parser')
	links = []
	for link in soup.find_all('a'):
		href = link.get('href')
		if href and urlparse(href).netloc == '':
			links.append(urljoin(url, href))
	return links



"""Routes to get chat history"""
import json
import traceback

from fastapi import APIRouter, HTTPException, Response, Depends, Request, status
from promptengineers.controllers import PromptController
from promptengineers.models.request import ReqBodyPromptSystem
from promptengineers.models.response import ResponsePromptSystemList, ResponsePromptSystem
from promptengineers.utils import JSONEncoder, logger

router = APIRouter()
TAG = "Prompt"

def get_controller(request: Request) -> PromptController:
	return PromptController(request=request)

#################################################
# List Chat Histories
#################################################
@router.get(
	"/prompt/system",
	tags=[TAG],
	response_model=ResponsePromptSystemList
)
async def index(
	page: int = 1,
	limit: int = 50,
	controller: PromptController = Depends(get_controller),
):
	"""List resources"""
	try:
		result = await controller.index(page, limit)
		# Format Response
		data = json.dumps({
			'prompts': result,
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[list_chat_histories]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create Chat History
#################################################
@router.post(
	"/prompt/system",
	tags=[TAG],
	response_model=ResponsePromptSystem
)
async def create(
	body: ReqBodyPromptSystem,
	controller: PromptController = Depends(get_controller)
):
	"""Creates resource"""
	try:
		result = await controller.create(body)
		# Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[create_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Show Chat History
#################################################
@router.get(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	response_model=ResponsePromptSystem,
)
async def show(
    prompt_id: str,
    controller: PromptController = Depends(get_controller),
):
	"""Retrieve resource"""
	try:
		result = await controller.show(prompt_id)

		# Format Response
		data = json.dumps({
			**result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[show_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Update Chat History
#################################################
@router.put(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	response_model=ResponsePromptSystem,
)
async def update(
	prompt_id: str,
	body: ReqBodyPromptSystem,
	controller: PromptController = Depends(get_controller),
):
	"""Update resource"""
	try:
		await controller.update(prompt_id, body)
		data = json.dumps({
			'message': 'Chat history updated successfully.'
		})
		# Format Response
		return Response(status_code=200, content=data)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[update_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Delete Chat History
#################################################
@router.delete(
	"/prompt/system/{prompt_id}",
	tags=[TAG],
	status_code=status.HTTP_204_NO_CONTENT,
)
async def delete(
	prompt_id: str,
	controller: PromptController = Depends(get_controller),
):
	"""Delete Resource"""
	try:
		await controller.delete(prompt_id)
		# Format Response
		return Response(status_code=204)
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[delete_chat_history]: %s\n%s", err, tb)
		raise HTTPException(status_code=404, detail=str(err)) from err


"""Response models for the API."""
from pydantic import BaseModel, Field

from promptengineers.models.response.http import (
    ResponseStatus,
    ResponseChat,
    ResponseAgentChat,
    ResponseAgentPluginsChat,
    ResponseVectorstoreChat,
    ResponseCreateVectorStore,
    ResponseListPineconeVectorStores,
)
from promptengineers.models.response.stream import (
    RESPONSE_STREAM_AGENT_CHAT,
    RESPONSE_STREAM_VECTORSTORE_CHAT,
    RESPONSE_STREAM_AGENT_PLUGINS_CHAT,
    RESPONSE_STREAM_CHAT,
)
from .prompt import ResponsePromptSystemList, ResponsePromptSystem


class ResponseFileLoader(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": "Vectorstore Created!",
                "vectorstore": "index_name",
            }
        }


class ResponseChatStream(BaseModel):
    """A message to send to the chatbot."""

    sender: str = Field(default="assistant")
    message: str = Field(default="Dialog started.")
    type: str = Field(default="stream")


class ResponseRetrieveVectorstores(BaseModel):
    """A message to send to the chatbot."""

    class Config:
        """A message to send to the chatbot."""

        json_schema_extra = {"example": {"vectorstores": ["formio.pkl", "bullmq.pkl"]}}


class ResponseRetrieveFiles(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "files": [
                    "A Prompt Pattern Catalog to Enhance Prompt Engineering with ChatGPT.pdf",
                    "ai-village.pdf",
                ]
            }
        }


class ResponseFileStorage(BaseModel):
    class Config:
        json_schema_extra = {
            "example": {
                "message": "File(s) Uploaded!",
                "bucket_name": "prompt-engineers-dev",
                "files": [
                    "formio-standard-procedure.pdf",
                    "formio-interview-questions.pdf",
                ],
            }
        }


__all__ = [
    "ResponseStatus",
    "ResponseChat",
    "ResponseAgentChat",
    "ResponseAgentPluginsChat",
    "ResponseVectorstoreChat",
    "RESPONSE_STREAM_AGENT_CHAT",
    "RESPONSE_STREAM_AGENT_PLUGINS_CHAT",
    "RESPONSE_STREAM_VECTORSTORE_CHAT",
    "RESPONSE_STREAM_CHAT",
    "ResponseCreateVectorStore",
    "ResponseListPineconeVectorStores",
    "ResponsePromptSystemList",
    "ResponsePromptSystem",
]



import os
import tempfile
from typing import List, Optional

from fastapi import File, HTTPException, UploadFile

from promptengineers.config import ACCESS_KEY_ID, BUCKET, ACCESS_SECRET_KEY
from promptengineers.config.test import TEST_USER_ID
from promptengineers.services.storage import StorageService


class StorageController:
	def __init__(self):
		# Initialize any necessary variables or objects here
		self.data = {}

	def save_files_to_bucket(self, files: Optional[List[UploadFile]] = File(...), directory: str = 'files'):
		try:
			s3client = StorageService(
				ACCESS_KEY_ID,
				ACCESS_SECRET_KEY
			)
		except Exception as err:
			raise HTTPException(status_code=500, detail=str(err)) from err

		# Create a temporary directory
		with tempfile.TemporaryDirectory() as tmpdirname:
			uploaded_files = []
			for file in files:
				if directory != 'files':
					ext = os.path.splitext(file.filename)[1][1:]
					if ext != 'pkl':
						raise HTTPException(
							status_code=400,
							detail=f"File [{file.filename}] can NOT be uploaded, only .pkl files are allowed."
						)
				## Save File To Temp Directory
				file_path = os.path.join(tmpdirname, file.filename)
				try:
					with open(file_path, 'wb') as reader:
						reader.write(file.file.read())  # write the file to the temporary directory
				except Exception as err:
					raise HTTPException(
						status_code=500,
						detail=f"Error saving files to temporary directory: {str(err)}"
					) from err
				# Now upload each file in the temporary directory to s3
				try:
					for file_name in os.listdir(tmpdirname):
						file_path = os.path.join(tmpdirname, file_name)
						s3_file_path = f'users/{TEST_USER_ID}/{directory}/{file_name}'  # Add the desired prefix
						s3client.upload_file(
							file_path,
							BUCKET,
							s3_file_path
						)
						uploaded_files.append(file_name)
				except Exception as err:
					raise HTTPException(
						status_code=500,
						detail=f"Error uploading files to S3: {str(err)}"
					) from err
		return {
			"files": uploaded_files
		}

	def retrieve_files_from_bucket(self, directory: str = 'files'):
		s3client = StorageService(
			ACCESS_KEY_ID,
			ACCESS_SECRET_KEY
		)
		files = s3client.retrieve_all_files(
			BUCKET,
			f'users/{TEST_USER_ID}/{directory}'
		)
		return {
			f'{directory}': [] if not files else files
		}

	##############################################################
	### Delete File From Bucket
	##############################################################
	def delete_file_from_bucket(self, file_name: str):
		try:
			## Delete File
			s3client = StorageService(
				ACCESS_KEY_ID,
				ACCESS_SECRET_KEY
			)
			s3client.delete_file(
				BUCKET,
				f'users/{TEST_USER_ID}/files/{file_name}'
			)
		except Exception as err:
			raise HTTPException(
				status_code=500,
				detail=f"Error deleting file from S3: {str(err)}"
            ) from err




from promptengineers.interfaces.repos import IUserRepo
from promptengineers.repos.user import UserRepo
from promptengineers.utils.validation import Validator
from promptengineers.strategies.vectorstores import PineconeStrategy, RedisStrategy

validator = Validator()

class VectorSearchProviderFactory:
	@staticmethod
	def choose(
		provider: ('redis', 'pinecone'),
		user_id: str,
		index_name: str,
		user_repo: IUserRepo = UserRepo(),
	):
		if provider == 'pinecone':
			required_keys = ['OPENAI_API_KEY', 'PINECONE_KEY', 'PINECONE_ENV', 'PINECONE_INDEX']
			tokens = user_repo.find_token(user_id, [*required_keys, 'PROMPTLAYER_API_KEY'])
			validator.validate_api_keys(tokens, required_keys)
			vectorstore_strategy = PineconeStrategy(
				openai_api_key=tokens.get(required_keys[0]),
				api_key=tokens.get(required_keys[1]),
				env=tokens.get(required_keys[2]),
				index_name=tokens.get(required_keys[3]),
				namespace=index_name,
			)

		if provider == 'redis':
			required_keys = ['OPENAI_API_KEY', 'REDIS_URL']
			tokens = user_repo.find_token(user_id, [*required_keys, 'PROMPTLAYER_API_KEY'])
			validator.validate_api_keys(tokens, required_keys)
			vectorstore_strategy = RedisStrategy(
				openai_api_key=tokens.get(required_keys[0]),
				redis_url=tokens.get(required_keys[1]),
				index_name=index_name,
			)
		return vectorstore_strategy


from typing import List
from pydantic import BaseModel, Field

class CopyPasteLoader(BaseModel):
	type: str = Field(...)
	text: str = Field(...)

class BlockchainLoader(BaseModel):
	type: str = Field(...)
	contract_address: str = Field(...)

class YoutubeLoader(BaseModel):
	type: str = Field(...)
	ytId: str = Field(...)

class TypeUrlLoader(BaseModel):
	type: str = Field(...)
	urls: List[str] = Field(...)



from typing import Union

from promptengineers.exceptions import ValidationException
from promptengineers.utils.errors import token_not_found_message


class Validator:

	def validate_api_keys(
		self,
		available_tokens: Union[dict, str],
		require_keys: list[str],
		exception_class=ValidationException
	) -> None:
		if isinstance(available_tokens, str):
			available_tokens = {require_keys[0]: available_tokens}
		error_message = token_not_found_message(available_tokens, require_keys)
		if error_message:
			raise exception_class(error_message)


"""Chat GPT Clone Prompt Template"""
from langchain.prompts import PromptTemplate

def get_system_template(system_message: str):
    prompt_template = f"""{system_message}
---
{{chat_history}}
User: {{question}}
Assistant: """
    template = PromptTemplate(
        template=prompt_template,
        input_variables=["chat_history", "question"]
    )
    return template

def get_retrieval_template(system_message: str):
    prompt_template = f"""{system_message}
---
{{context}}
---
{{chat_history}}
User: {{question}}
Assistant: """
    template = PromptTemplate(
        template=prompt_template,
        input_variables=["context", "question", "chat_history"]
    )
    return template


"""Chain Service"""
from langchain.schema import SystemMessage
from langchain.agents import (AgentType, initialize_agent,
							AgentExecutor, OpenAIFunctionsAgent,
							load_tools)
from langchain.memory import ConversationBufferMemory
from langchain.prompts import MessagesPlaceholder
from langchain.tools import AIPluginTool
from langchain.callbacks.base import BaseCallbackHandler
from langchain.chains import (
    ConversationChain,
	ConversationalRetrievalChain,
	LLMChain
)
from langchain.chains.chat_vector_db.prompts import QA_PROMPT
from langchain.chains.question_answering import load_qa_chain
from langchain.prompts import (
    ChatPromptTemplate,
    SystemMessagePromptTemplate,
    MessagesPlaceholder,
    HumanMessagePromptTemplate,
    PromptTemplate
)
from promptengineers.strategies.vectorstores import VectorstoreContext
from promptengineers.config import APP_ENV
from promptengineers.config.tools import AVAILABLE_TOOLS
from promptengineers.utils.chains import get_chat_history, filter_tools
from promptengineers.utils.prompts import get_system_template

from langchain.agents.openai_functions_agent.agent_token_buffer_memory import (
    AgentTokenBufferMemory,
)
from langchain.agents.agent_toolkits import create_conversational_retrieval_agent, create_retriever_tool



class ChainService:
	"""Chain Service"""
	def __init__(self, model):
		self.model = model

	def condense_question(self, system_message):
		"""Condense a question into a single sentence."""
		return LLMChain(
			llm=self.model,
			prompt=get_system_template(system_message),
		)

	def collect_docs(self, system_message):
		"""Collect documents from the vectorstore."""
		return load_qa_chain(
			self.model,
			chain_type='stuff',
			prompt=get_system_template(system_message)
		)

	def create_executor(
		self,
		content,
		tools,
		chat_history,
		verbose=True if APP_ENV == 'local' or APP_ENV == 'development' else False,
		return_messages = True,
		callbacks = [],
		return_intermediate_steps = True
	):
		# memory = ConversationBufferMemory(
		# 	memory_key="chat_history", 
		# 	return_messages=return_messages, 
		# 	output_key='output'
		# )
		memory = AgentTokenBufferMemory(memory_key="chat_history", llm=self.model, return_messages=return_messages)
		system_message = SystemMessage(content=content)
		prompt = OpenAIFunctionsAgent.create_prompt(
			system_message=system_message,
			extra_prompt_messages=[MessagesPlaceholder(variable_name="chat_history")]
		)
		if len(chat_history) > 0:
			for message in chat_history:
				if message[0] and message[1]:
					memory.chat_memory.add_user_message(message[0])
					memory.chat_memory.add_ai_message(message[1])
				else:
					memory.chat_memory.add_user_message(message[0])
		agent = OpenAIFunctionsAgent(llm=self.model, tools=tools, prompt=prompt)
		return AgentExecutor(
			agent=agent, 
			tools=tools,
			memory=memory, 
			verbose=verbose, 
			callbacks=callbacks, 
			return_intermediate_steps=return_intermediate_steps
		)
	
	# def conversation_retrieval(
	# 	self,
	# 	vectorstore,
	# 	system_message,
	# 	chat_history,
	# 	callbacks = []
	# ):
	# 	"""Retrieve a conversation."""
	# 	tool = create_retriever_tool(
	# 		vectorstore.as_retriever(),
	# 		"search_formio_docs",
	# 		"Searches and returns documents regarding Form.io.",
	# 	)
	# 	tools = [tool]
	# 	system_message = SystemMessage(content=system_message)
	# 	agent_executor = create_conversational_retrieval_agent(
	# 		self.model, 
	# 		tools, 
	# 		system_message=system_message, 
	# 		verbose=True,
	# 		# callbacks=callbacks
	# 	)
	# 	return agent_executor

	def conversation_retrieval(
		self,
		system_message,
		chat_history,
		vectorstore,
		verbose=True if APP_ENV == 'local' or APP_ENV == 'development' else False,
	):
		"""Retrieve a conversation."""
		tool = create_retriever_tool(
			vectorstore.as_retriever(),
			"search_docs",
			"Searches and returns documents. It is a requirement to use this for every query.",
		)
		tools = [tool]
		system_message = SystemMessage(content=system_message)
		prompt = OpenAIFunctionsAgent.create_prompt(
			system_message=system_message,
			extra_prompt_messages=[MessagesPlaceholder(variable_name="chat_history")],
		)
		agent = OpenAIFunctionsAgent(llm=self.model, tools=tools, prompt=prompt)
		memory = AgentTokenBufferMemory(memory_key="chat_history", llm=self.model)
		for message in chat_history:
			if message[0] and message[1]:
				memory.chat_memory.add_user_message(message[0])
				memory.chat_memory.add_ai_message(message[1])
			else:
				memory.chat_memory.add_user_message(message[0])
		agent_executor = AgentExecutor(
			agent=agent,
			tools=tools,
			memory=memory,
			verbose=verbose,
			return_intermediate_steps=True,
		)
		return agent_executor

	# def conversation_retrieval(
	# 	self,
	# 	vectorstore,
	# 	system_message,
	# 	chat_history,
	# 	callbacks = []
	# ):
	# 	"""Retrieve a conversation."""
	# 	memory = ConversationSummaryBufferMemory(llm=self.model, memory_key="chat_history", return_messages=True)
	# 	for message in chat_history:
	# 		if message[0] and message[1]:
	# 			memory.chat_memory.add_user_message(message[0])
	# 			memory.chat_memory.add_ai_message(message[1])
	# 		else:
	# 			memory.chat_memory.add_user_message(message[0])
	# 	return ConversationalRetrievalChain.from_llm(
	# 		condense_question_llm=self.model,
	# 		retriever=vectorstore.as_retriever(),
	# 		memory=memory,
	# 		combine_docs_chain=self.collect_docs(system_message),
	# 		get_chat_history=get_chat_history,
	# 		verbose=True,
	# 		callbacks=callbacks
	# 	)

	# def conversation_retrieval(
	# 	self,
	# 	vectorstore,
	# 	system_message,
	# 	chat_history,
	# 	callbacks = []
	# ):
	# 	"""Retrieve a conversation."""
	# 	memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
	# 	if len(chat_history) > 0:
	# 		for message in chat_history:
	# 			if message[0] and message[1]:
	# 				memory.chat_memory.add_user_message(message[0])
	# 				memory.chat_memory.add_ai_message(message[1])
	# 			else:
	# 				memory.chat_memory.add_user_message(message[0])
	# 	return ConversationalRetrievalChain(
	# 		question_generator=self.condense_question(system_message),
	# 		retriever=vectorstore.as_retriever(),
	# 		memory=memory,
	# 		combine_docs_chain=self.collect_docs(system_message),
	# 		get_chat_history=get_chat_history,
	# 		verbose=True,
	# 		callbacks=callbacks
	# 	)

	# def agent_search(self, tools, chat_history):
	# 	"""Agent search."""
	# 	memory = ConversationBufferMemory(memory_key="chat_history", return_messages=True)
	# 	if len(chat_history) > 0:
	# 		for message in chat_history:
	# 			if message[0] and message[1]:
	# 				memory.chat_memory.add_user_message(message[0])
	# 				memory.chat_memory.add_ai_message(message[1])
	# 			else:
	# 				memory.chat_memory.add_user_message(message[0])
	# 	return initialize_agent(
	# 		tools,
	# 		self.model,
	# 		agent=AgentType.CHAT_CONVERSATIONAL_REACT_DESCRIPTION or AgentType.CHAT_ZERO_SHOT_REACT_DESCRIPTION,
	# 		verbose=True,
	# 		memory=memory,
	# 		get_chat_history=get_chat_history
	# 	)

	def agent_with_tools(
			self, 
			system_message: str, 
			chat_history, 
			tools: list[str] = None,
			available_tools: dict[str, any] = None,
			vectorstore: VectorstoreContext = None,
			plugins: list[str] = None,
			callbacks: list[BaseCallbackHandler] = None
		):
		"""Agent search."""
		filtered_tools = filter_tools(tools or [], available_tools or AVAILABLE_TOOLS)
		
		## Add docs tool
		if vectorstore:
			docs_tool = create_retriever_tool(
				vectorstore.as_retriever(),
				"search_docs",
				"""
				Searches and returns relevant documents. It is a requirement to use this for every query.
				Rewrite the user to be a detailed question.
				""",
			)
			filtered_tools.append(docs_tool)

		## Add plugins
		if plugins and len(plugins) > 0:
			loaded_tools = load_tools(["requests_all"])
			for tool in plugins:
				tool = AIPluginTool.from_plugin_url(tool)
				loaded_tools += [tool]
			filtered_tools += loaded_tools
		## Create agent
		agent_executor = self.create_executor(system_message, filtered_tools, chat_history, callbacks=callbacks)
		return agent_executor

	# def agent_with_plugins(self, plugins, system_message, chat_history, callbacks=[]):
	# 	"""Agent Plugins."""
	# 	loaded_tools = load_tools(["requests_all"])
	# 	for tool in plugins:
	# 		tool = AIPluginTool.from_plugin_url(tool)
	# 		loaded_tools += [tool]
	# 	agent_executor = self.create_executor(system_message, loaded_tools, chat_history, callbacks=callbacks)
	# 	return agent_executor

	def conversation(self):
		prompt_template = ChatPromptTemplate.from_messages(
			[
				MessagesPlaceholder(variable_name="context"),
				HumanMessagePromptTemplate.from_template("{input}")
			]
		)
		memory = ConversationBufferMemory(return_messages=True, memory_key="context")
		llm_chain = ConversationChain(llm=self.model, prompt=prompt_template, memory=memory, verbose=False)

		return llm_chain


import ujson
import traceback

from fastapi import APIRouter, Depends, Request, Response, HTTPException
from fastapi.responses import StreamingResponse

from promptengineers.controllers import ChatController, AuthController
from promptengineers.exceptions import ValidationException
from promptengineers.factories.provider import VectorSearchProviderFactory
from promptengineers.models.request import (ReqBodyChat, ReqBodyAgentChat,
									ReqBodyVectorstoreChat, ReqBodyAgentPluginsChat)
from promptengineers.models.response import (ResponseChat, ResponseAgentChat, ResponseVectorstoreChat,
									ResponseAgentPluginsChat, RESPONSE_STREAM_AGENT_PLUGINS_CHAT,
									RESPONSE_STREAM_AGENT_CHAT, RESPONSE_STREAM_VECTORSTORE_CHAT,
									RESPONSE_STREAM_CHAT)
from promptengineers.repos.user import UserRepo
from promptengineers.strategies.vectorstores import VectorstoreContext
from promptengineers.utils import logger
from promptengineers.utils.chains import format_agent_actions
from promptengineers.utils.validation import Validator

user_repo = UserRepo()
validator = Validator()
auth_controller = AuthController()
router = APIRouter()
TAG = "Chat"

def get_controller(request: Request) -> ChatController:
	return ChatController(request=request)

#################################################
# ChatGPT
#################################################
@router.post(
	"/chat",
	tags=[TAG],
	response_model=ResponseChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_CHAT
				}
			}
		}
	},
)
async def chat(
	body: ReqBodyChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat endpoint."""
	try:
		logger.debug('[POST /chat] Query: %s', str(body))
		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = chat_controller.langchain_http_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature
			)
			data = ujson.dumps({
				'message': result,
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		return StreamingResponse(
			chat_controller.langchain_stream_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature
			),
			headers={
				"Cache-Control": "no-cache",
				"Connection": "keep-alive",
				"Content-Type": "text/event-stream",
			}
		)
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.files.list_files] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err



#################################################
# Langchain Agent
#################################################
@router.post(
	"/chat/agent",
	tags=[TAG],
	response_model=ResponseAgentChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_AGENT_CHAT
				}
			}
		}
	},
)
async def agent(
	body: ReqBodyAgentChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat endpoint."""
	try:
		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = chat_controller.langchain_http_agent_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				tools=body.tools
			)
			data = ujson.dumps({
				'message': result['output'],
				'actions': format_agent_actions(result['intermediate_steps']),
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat/agent] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		return StreamingResponse(
			chat_controller.langchain_stream_agent_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				tools=body.tools
			),
			headers={
				"Cache-Control": "no-cache",
				"Connection": "keep-alive",
				"Content-Type": "text/event-stream",
			}
		)
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.chat.agent] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err

#################################################
# Langchain Vectorstore Route
#################################################
@router.post(
	"/chat/vectorstore",
	tags=[TAG],
	response_model=ResponseVectorstoreChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_VECTORSTORE_CHAT
				}
			}
		}
	},
)
async def vector_search(
	request: Request,
	body: ReqBodyVectorstoreChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat Vectorstore endpoint."""
	try:
		# Log Context Details
		logger.debug('[POST /chat/vectorstore] Query: %s', str(body))

		# Retrieve User Tokens
		user_id = getattr(request.state, "user_id", None)

		# Retreve Vectorstore
		vectorstore_strategy = VectorSearchProviderFactory.choose(
			provider=body.provider,
			user_id=user_id,
			index_name=body.vectorstore
		)
		vectostore_service = VectorstoreContext(vectorstore_strategy)
		vectorstore = vectostore_service.load()

		# Check if the retrieved file is empty
		if not vectorstore:
			raise HTTPException(
				status_code=404,
				detail=f"Vectorstore {body.vectorstore} not found"
			)

		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = chat_controller.langchain_http_vectorstore_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				vectorstore=vectorstore,
			)
			data = ujson.dumps({
				'message': result,
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat/vectorstore] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		# Process Query
		return StreamingResponse(
			chat_controller.langchain_stream_vectorstore_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				vectorstore=vectorstore,
			),
			media_type="text/event-stream"
		)
	except ValidationException as err:
		logger.warning("[routes.chat.vector_search] ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("[routes.chat.vector_search] HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.chat.vector_search] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err



"""Chat History Request Models"""
from typing import Optional

from promptengineers.models.request.chat import ReqBodyChat

class ReqBodyChatHistory(ReqBodyChat): # pylint: disable=too-few-public-methods
	"""A message to send to the chatbot."""
	title: Optional[str] = None

	class Config: # pylint: disable=too-few-public-methods
		"""A message to send to the chatbot."""
		json_schema_extra = {
			"example": {
				"title": "World Series 2001 Chatbot",
				"model": "gpt-3.5-turbo",
				"temperature": 0.8,
				"stream": False,
				"messages": [
					{"role": "system", "content": "You are a helpful assistant."},
					{"role": "user", "content": 'Who won the 2001 world series?'},
					{"role": "assistant", "content": 'The arizona diamondbacks won the 2001 world series.'},
					{"role": "user", "content": 'Who were the pitchers?'},
				]
			}
		}

class ReqBodyListChatHistory(ReqBodyChatHistory): # pylint: disable=too-few-public-methods
	"""A message to send to the chatbot."""

	class Config: # pylint: disable=too-few-public-methods
		"""A message to send to the chatbot."""
		json_schema_extra = {
			"example": {
				"chats": [
					{
						"_id": "653e147a126c8e67d951fd20",
						"title": "World Series 2001 Chatbot",
						"model": "gpt-3.5-turbo",
						"temperature": 0.8,
						"stream": False,
						"messages": [
							{"role": "system", "content": "You are a helpful assistant."},
							{"role": "user", "content": 'Who won the 2001 world series?'},
							{"role": "assistant", "content": 'The arizona diamondbacks won the 2001 world series.'},
							{"role": "user", "content": 'Who were the pitchers?'},
						],
						"functions": [],
						"vectorstore": "",
						"user_id": "63f0962f9a09c84c98ab6caf",
						"created_at": 1698523723,
						"updated_at": 1698562747
					},
					{
						"_id": "6539f7f3126c8e67d951fc77",
						"temperature": 0.9,
						"model": "gpt-3.5-turbo",
						"messages": [
							{
								"role": "system",
								"content": "PERSONA:\nImagine you super intelligent AI assistant that is an expert on the context.\n\nINSTRUCTION:\nUse the following pieces of context to answer the question at the end. If you don't know the answer or if the required code is not present, just say that you don't know, and don't try to make up an answer. \n\nOUTPUT FORMAT RULES:\nCode snippets should be wrapped in triple backticks, along with the language name for proper formatting,  if applicable. This also includes yaml and Dockerfile's. If showing how to install dependencies like npm, pip, cargo, etc use the bash backticks.\nExample:\n```python\nprint(\"Hello World!\")\n```"
							},
							{
								"role": "user",
								"content": "Can you summarize the readme on chat-extension repository "
							},
							{
								"role": "assistant",
								"content": "I'm sorry, but I couldn't find the README for the chat-extension repository."
							}
						],
						"tools": [
							"github_get_repository",
							"github_list_organization_repos",
							"github_issues_assigned_to_auth_user",
							"github_repo_issue",
							"github_repo_issue_create",
							"github_repo_issue_update",
							"github_repo_pull_create"
						],
						"functions": [],
						"vectorstore": "",
						"user_id": "63f0962f9a09c84c98ab6caf",
						"created_at": 1698297843,
						"updated_at": 1698297843
					}
				]
			}
		}



from promptengineers.models.prompt import PromptSystem

class ReqBodyPromptSystem(PromptSystem): # pylint: disable=too-few-public-methods

	class Config: # pylint: disable=too-few-public-methods
		"""Prompt System Template"""
		json_schema_extra = {
			"example": {
				"title": "Helpful Assistant Prompt",
				"system": "You are a helpful {ASSISTANT_TYPE} assistant. You have have been given context to {INDEX_NAME} to answer user questions.",
				"variables": {
					"ASSISTANT_TYPE": 'document retreival',
					"INDEX_NAME": 'Stripe API docs'
                }
			}
		}


# import pickle
from abc import ABC, abstractmethod
# from io import BytesIO

from langchain.embeddings.openai import OpenAIEmbeddings

from promptengineers.services.pinecone import PineconeService
from promptengineers.services.redis import RedisService


# Define the strategy interface
class VectorstoreStrategy(ABC):
    @abstractmethod
    def load(self):
        pass

#########################################################################
## FAISS Strategy
#########################################################################
# class FaissStrategy(VectorstoreStrategy):
#     def __init__(self, _s3_access_key, _s3_secret_key, bucket_name, path):
#         self.s3_access_key = _s3_access_key
#         self.s3_secret_key = _s3_secret_key
#         self.bucket_name = bucket_name
#         self.path = path

#     def load(self):
#         f = BucketService(
#             self.s3_access_key,
#             self.s3_secret_key
#         ).retrieve_file(
#             bucket=self.bucket_name,
#             path=self.path
#         )
#         with BytesIO(f.read()) as file:
#             vectorstore = pickle.load(file)
#         return vectorstore

#########################################################################
## Pinecone Strategy
#########################################################################
class PineconeStrategy(VectorstoreStrategy):
    def __init__(
        self,
        openai_api_key: str,
        api_key: str,
        env: str,
        namespace: str,
        index_name: str,
    ):
        self.api_key = api_key
        self.openai_api_key = openai_api_key
        self.env = env
        self.namespace = namespace
        self.index_name = index_name

    def load(self):
        embeddings = OpenAIEmbeddings(openai_api_key=self.openai_api_key)
        pinecone_service = PineconeService(
			api_key=self.api_key,
			env=self.env,
			index_name=self.index_name,
		)
        return pinecone_service.from_existing(
            embeddings,
            namespace=self.namespace
        )

#########################################################################
## Pinecone Strategy
#########################################################################
class RedisStrategy(VectorstoreStrategy):
    def __init__(
        self,
        openai_api_key: str,
        redis_url: str,
        index_name: str,
        index_schema: dict = None,
    ):
        self.redis_url = redis_url
        self.openai_api_key = openai_api_key
        self.index_name = index_name
        self.index_schema = index_schema

    def load(self):
        redis_service = RedisService(
			redis_url=self.redis_url,
			index_name=self.index_name,
			openai_api_key=self.openai_api_key,
		)
        return redis_service.from_existing(
            schema=self.index_schema
        )

#########################################################################
## Strategy Context
#########################################################################
class VectorstoreContext:
    def __init__(self, strategy: VectorstoreStrategy):
        self.strategy = strategy

    def set_strategy(self, strategy: VectorstoreStrategy):
        self.strategy = strategy

    def load(self):
        return self.strategy.load()



"""Stream Utils"""
from typing import Awaitable
import asyncio
import ujson
import openai

from promptengineers.utils import logger


async def wrap_done(fn_name: Awaitable, event: asyncio.Event):
    """Wrap an awaitable with a event to signal when it's done or an exception is raised."""
    try:
        await fn_name
    except asyncio.CancelledError:
        pass
    except openai.APIError as error:
        print(f"Caught API error: {error}")
    finally:
        # Signal the aiter to stop.
        event.set()

def token_stream(token: str = None, action_type: str = None):
    """ Use server-sent-events to stream the response"""
    if not token and not action_type:
        data = {
            'sender': 'assistant',
            'message': "",
            'type': 'end'
        }
        logger.debug('[utils.stream.token_stream] End: %s', str(data))
    elif action_type == 'tool':
        data = {
            'sender': 'assistant',
            'message': token,
            'type': 'tool'
        }
        logger.debug('[utils.stream.token_stream] Action: %s', str(data))
    elif action_type == 'log':
        data = {
            'sender': 'assistant',
            'message': token,
            'type': 'log'
        }
        logger.debug('[utils.stream.token_stream] Log: %s', str(data))
    else:
        data = {
            'sender': 'assistant',
            'message': token,
            'type': 'stream'
        }
        logger.debug('[utils.stream.token_stream] Token: %s', str(data))
    return f"data: {ujson.dumps(data)}\n\n"


from datetime import datetime, timedelta

from langchain.chains import LLMMathChain
from langchain.agents import Tool, tool
from langchain.callbacks.streaming_stdout_final_only import FinalStreamingStdOutCallbackHandler

from promptengineers.config import OPENAI_API_KEY
from promptengineers.strategies.llms import ModelContext, OpenAIStrategy

model_service = ModelContext(strategy=OpenAIStrategy(api_key=OPENAI_API_KEY))
llm = model_service.chat(
    model_name='gpt-3.5-turbo-16k',
    temperature=0.0,
    callbacks=[FinalStreamingStdOutCallbackHandler()]
)
llm_math = LLMMathChain(llm=llm)

math_tool = Tool(
    name='Calculator',
    func=llm_math.run,
    description='Useful for when you need to answer questions about math.'
)

@tool
def get_word_length(word: str) -> int:
    """Returns the length of a word."""
    return len(word)

@tool
def fetch_start_and_end(start: str = None, end: str = None) -> tuple:
    """
    Convert start and end datetime strings to Unix timestamps in milliseconds.

    Args:
        start (str, optional): Start datetime in the format "YYYY-MM-DD HH:MM:SS". Defaults to current time.
        end (str, optional): End datetime in the same format. Defaults to one hour after start.

    Returns:
        tuple: Unix timestamps for start and end in milliseconds.
    """
    # Get the current time
    current_time = datetime.now()

    # Convert start string to datetime object
    if not start:
        start_datetime = current_time
    else:
        start_datetime = datetime.strptime(start, "%Y-%m-%d %H:%M:%S")

    # If no end time is provided, set it to one hour after start
    if not end:
        end_datetime = start_datetime + timedelta(hours=1)
    else:
        end_datetime = datetime.strptime(end, "%Y-%m-%d %H:%M:%S")

    # Convert datetime objects to Unix timestamp in milliseconds
    start_unix = int(start_datetime.timestamp() * 1000)
    end_unix = int(end_datetime.timestamp() * 1000)

    return (start_unix, end_unix)





import os

from fastapi import FastAPI, Depends, HTTPException, status, Request
from fastapi.security import HTTPBasic, HTTPBasicCredentials

from promptengineers.config.test import TEST_USER_ID

security = HTTPBasic()


class AuthController:
	def __init__(self) -> None:
		self.users_db = self.load_users_from_env()

	def load_users_from_env(self):
		users = {}
		for key, value in os.environ.items():
			if key.startswith("USER_"):
				username = key[5:]  # Remove 'USER_' prefix
				users[username] = value
		return users

	def get_current_user(
		self,
		request: Request,
		credentials: HTTPBasicCredentials = Depends(security)
	):
		user = self.users_db.get(credentials.username)
		if user is None or user != credentials.password:
			raise HTTPException(
				status_code=status.HTTP_401_UNAUTHORIZED,
				detail="Invalid credentials",
				headers={"WWW-Authenticate": "Basic"},
			)
		else:
			request.state.user_id = TEST_USER_ID
			return request


"""Routes to get chat history"""
import json
import traceback

from fastapi import APIRouter, HTTPException, Response, Depends, Request, status
from promptengineers.controllers import HistoryController
from promptengineers.models.request import ReqBodyListChatHistory, ReqBodyChatHistory
from promptengineers.utils import JSONEncoder, logger

router = APIRouter()
TAG = "Chat"

def get_controller(request: Request) -> HistoryController:
	return HistoryController(request=request)

#################################################
# List Chat Histories
#################################################
@router.get(
	"/chat/history",
	tags=[TAG],
	response_model=ReqBodyListChatHistory
)
async def list_chat_histories(
	page: int = 1,
	limit: int = 50,
	controller: HistoryController = Depends(get_controller),
):
	"""Creates a chat history"""
	try:
		result = await controller.index(page, limit)
		# Format Response
		data = json.dumps({
			'chats': result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[list_chat_histories]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create Chat History
#################################################
@router.post(
	"/chat/history",
	tags=[TAG],
	response_model=ReqBodyChatHistory
)
async def create_chat_history(controller: HistoryController = Depends(get_controller)):
	"""Creates a chat history"""
	try:
		result = await controller.create()
		# Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[create_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Show Chat History
#################################################
@router.get(
	"/chat/history/{chat_id}",
	tags=[TAG],
	response_model=ReqBodyChatHistory,
)
async def show_chat_history(
    chat_id: str,
    controller: HistoryController = Depends(get_controller),
):
	"""Creates a chat history"""
	try:
		result = await controller.show(chat_id)

		# Format Response
		data = json.dumps({
			**result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[show_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Update Chat History
#################################################
@router.put(
	"/chat/history/{chat_id}",
	tags=[TAG],
	response_model=ReqBodyChatHistory,
)
async def update_chat_history(
	chat_id: str,
	controller: HistoryController = Depends(get_controller),
):
	"""Update chat history"""
	try:
		await controller.update(chat_id)
		data = json.dumps({
			'message': 'Chat history updated successfully.'
		})
		# Format Response
		return Response(status_code=200, content=data)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[update_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Delete Chat History
#################################################
@router.delete(
	"/chat/history/{chat_id}",
	tags=[TAG],
	status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_chat_history(
	chat_id: str,
	controller: HistoryController = Depends(get_controller),
):
	"""Deletes chat history"""
	try:
		await controller.delete(chat_id)
		# Format Response
		return Response(status_code=204)
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[delete_chat_history]: %s\n%s", err, tb)
		raise HTTPException(status_code=404, detail=str(err)) from err


from cryptography.fernet import Fernet

from promptengineers.config import APP_SECRET

def encrypt(data: str) -> str:
	return Fernet(APP_SECRET).encrypt(data.encode()).decode()

def decrypt(data: str) -> str:
	return Fernet(APP_SECRET).decrypt(data.encode()).decode()


"""Service for interacting with the history collection in MongoDB."""
import time
import motor.motor_asyncio

class MongoService:
	"""Service for interacting with the history collection in MongoDB."""
	def __init__(self, host: str, db: str, collection: str):
		client = motor.motor_asyncio.AsyncIOMotorClient(host)
		self.database = client[db]
		self.collection = self.database[collection]

	async def ensure_ttl_index(self):
		"""Ensure that the TTL index is set on the created_at field."""
		await self.collection.create_index("created_at", expireAfterSeconds=3600 * 24 * 7)

	async def list_docs(self, params: dict, limit=None, page=1, projection={'user_id': 0}) -> list:
		"""List documents in the collection with pagination."""
		cursor = self.collection.find(params, projection).sort("updated_at", -1)

		if limit:
			skip = (page - 1) * limit
			cursor = cursor.skip(skip).limit(limit)
		else:
			limit = 1000

		documents = await cursor.to_list(length=limit)
		return documents

	async def create(self, document: dict) -> str:
		"""Insert a new document into the collection."""
		document['created_at'] = int(time.time())
		document['updated_at'] = int(time.time())
		result = await self.collection.insert_one(dict(document))
		return {
			'_id': str(result.inserted_id)
		}

	async def read_one(self, params: dict, projection={'user_id': 0}) -> dict:
		"""Find and return a single document from the collection."""
		document = await self.collection.find_one(params, projection)
		return document

	async def update_one(self, params: dict, update: dict) -> bool:
		"""Update a single document in the collection."""
		update['updated_at'] = int(time.time())
		result = await self.collection.update_one(
			params,
			{'$set': update}
		)
		return result.modified_count > 0

	async def delete_one(self, params: dict) -> bool:
		"""Delete a single document from the collection."""
		result = await self.collection.delete_one(params)
		return result.deleted_count > 0



from enum import Enum

class OpenAIModels(Enum):
    GPT_3_5_TURBO = 'gpt-3.5-turbo'
    GPT_3_5_TURBO_16K = 'gpt-3.5-turbo-1106'
    GPT_3_5_TURBO_MARCH = 'gpt-3.5-turbo-0301'
    GPT_4 = 'gpt-4'
    GPT_4_MARCH = 'gpt-4-0314'
    GPT_4_32K = 'gpt-4-32k'
    GPT_4_32K_MARCH = 'gpt-4-32k-0314'
    GPT_4_TURBO = 'gpt-4-1106-preview'
    GPT_4_VISION = 'gpt-4-vision-preview'

class OllamaModels(Enum):
    LLAMA_2 = 'llama2'
    CODE_LLAMA = 'codellama'
    VICUNA = 'vicuna'
    MISTRAL = 'mistral'

ACCEPTED_OPENAI_MODELS = {
    OpenAIModels.GPT_3_5_TURBO.value,
    OpenAIModels.GPT_3_5_TURBO_16K.value,
    OpenAIModels.GPT_4.value,
    OpenAIModels.GPT_4_32K.value,
    OpenAIModels.GPT_4_TURBO.value
}

ACCEPTED_OLLAMA_MODELS = {
    OllamaModels.LLAMA_2.value,
    OllamaModels.CODE_LLAMA.value,
    OllamaModels.VICUNA.value,
    OllamaModels.MISTRAL.value
}




from langchain.callbacks import get_openai_callback
from langchain.embeddings import OpenAIEmbeddings
from langchain.text_splitter import CharacterTextSplitter, RecursiveCharacterTextSplitter
from langchain.vectorstores import FAISS

from promptengineers.utils import logger

def split_docs(
    documents,
    chunk_size: int = 1000,
    chunk_overlap: int = 100,
    recursive: bool = True,
):
    ## Text Spliter
    if recursive:
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )
    else:
        text_splitter = CharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )
    docs = text_splitter.split_documents(documents)
    return docs

def create_faiss_vectorstore(
    docs,
    chunk_size: int = 1000,
    chunk_overlap: int = 0
):
    """Load the vectorstore."""
    with get_openai_callback() as callback:
        logger.info('[utils.retrieval.create_vectorstore] Tokens: %s', callback.total_tokens)
        embeddings = OpenAIEmbeddings(max_retries=2)
        return FAISS.from_documents(
            split_docs(docs, chunk_size, chunk_overlap),
            embeddings
        )



from bson.objectid import ObjectId
# from fastapi import Request

from promptengineers.interfaces.repos import IUserRepo
from promptengineers.repos.user import UserRepo
from promptengineers.services.mongo import MongoService
from promptengineers.models.request import ReqBodyPromptSystem

class PromptController:
	def __init__(
		self, 
		# request: Request = None, 
		request = None,
		user_repo: IUserRepo = None, 
		db_name: str = None,
		col_name: str = None
	):
		self.request = request
		self.user_id = getattr(request.state, "user_id", None)
		self.user_repo = user_repo or UserRepo()
		self.prompt_service = MongoService(
			host=self.user_repo.find_token(self.user_id, 'MONGO_CONNECTION'),
			db=db_name or self.user_repo.find_token(self.user_id, 'MONGO_DB_NAME'),
			collection=col_name or 'prompts'
		)

	##############################################################
	### Create Chat History
	##############################################################
	async def index(self, page: int = 1, limit: int = 10):
		result = await self.prompt_service.list_docs(
			{'user_id': ObjectId(self.user_id)},
			limit,
			page
		)
		return result

	##############################################################
	### Create Chat History
	##############################################################
	async def create(self, body: ReqBodyPromptSystem):
		body = await self.request.json()
		body['user_id'] = ObjectId(self.user_id)
		result = await self.prompt_service.create(dict(body))
		return result

	##############################################################
	### Update Chat History
	##############################################################
	async def show(self, chat_id: str):
		result = await self.prompt_service.read_one(
			{'_id': ObjectId(chat_id), 'user_id': ObjectId(self.user_id)}
		)
		return result


	##############################################################
	### Update Chat History
	##############################################################
	async def update(self, chat_id: str, body: ReqBodyPromptSystem):
		body = await self.request.json()
		result = await self.prompt_service.update_one(
			{'_id': ObjectId(chat_id), 'user_id': ObjectId(self.user_id)},
			dict(body)
		)
		return result

	##############################################################
	### Delete Chat History
	##############################################################
	async def delete(self, chat_id: str):
		result = await self.prompt_service.delete_one(
			{'_id': ObjectId(chat_id), 'user_id': ObjectId(self.user_id)}
		)
		return result


"""Utilites for Chains"""
# from langchain.schema.agent import AgentActionMessageLog

def get_chat_history(inputs: tuple) -> str:
    """Formats the chat history into a readable format for the chatbot"""
    res = []
    for human, assistant in inputs:
        res.append(f"Human: {human}\nAssistant: {assistant}")
    return "\n".join(res)

def filter_tools(keys, dictionary):
    """
    Fetches values from the dictionary based on provided keys.

    Args:
    - dictionary (dict): The source dictionary.
    - keys (list): List of keys to fetch values for.

    Returns:
    - list: List of values corresponding to the provided keys.
    """
    return [dictionary.get(key) for key in keys]

def format_agent_actions(steps: list[tuple]) -> list[dict]:
    return [
        {"tool": step[0].tool, "tool_input": step[0].tool_input, "log": step[0].log}
        for step in steps
    ]




import json
import traceback
from typing import List

from fastapi import (APIRouter, File, HTTPException, Response,
					UploadFile, status, Depends)

from promptengineers.config import ACCESS_KEY_ID, BUCKET, ACCESS_SECRET_KEY
from promptengineers.config.test import TEST_USER_ID
from promptengineers.controllers import StorageController, AuthController
from promptengineers.models.response import ResponseFileStorage, ResponseRetrieveFiles
from promptengineers.services.storage import StorageService
from promptengineers.utils import logger

TAG = "Storage"
router = APIRouter()
auth_controller = AuthController()

#################################################
## List Files
#################################################
@router.get(
	"/files",
	response_model=ResponseRetrieveFiles,
	tags=[TAG],
)
async def list_files():
	try:
		result = StorageController().retrieve_files_from_bucket()
		## Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.files.list_files] Exception: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
## Add files to storage
#################################################
@router.post(
	"/files",
	response_model=ResponseFileStorage,
	tags=[TAG],
)
async def save_files(
	files: List[UploadFile] = File(...)
):
	try:
		result = StorageController().save_files_to_bucket(files)
		## Format Response
		data = json.dumps({
			'message': 'File(s) Uploaded!',
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except Exception as err:
		logger.error(err)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

######################################
##      Delete Vector Store
######################################
@router.delete(
	"/files",
	status_code=status.HTTP_204_NO_CONTENT,
	tags=[TAG],
)
async def delete_file(
	prefix: str,
):
	try:
		## Delete File
		s3client = StorageService(
			ACCESS_KEY_ID,
			ACCESS_SECRET_KEY
		)
		s3client.delete_file(
			BUCKET,
			f'users/{TEST_USER_ID}/files/{prefix}'
		)
		return Response(status_code=204)
	except Exception as err:
		raise HTTPException(
			status_code=404,
			detail=str(err)
		) from err



import os
from setuptools import setup, find_packages

VERSION = os.environ.get('VERSION', '0.0.1') 

setup(
    name='promptengineers',
    version=VERSION,
    packages=find_packages(exclude=["tests*", ".github*", "scripts*"]),
    install_requires=[
        'ujson',
    ],
    extras_require={
        'fastapi': [
            'fastapi',
            'uvicorn',
            ## History
            'motor', 'pymongo', 'cryptography',
            ## Storage
            'minio', 'python-multipart',
            ## Chat & Retrieval
            'langchain', 
            'openai', 
            'redis', 
            'pinecone-client', 
            'youtube-transcript-api', 
            'pypdf', 
            'numexpr', 
            'tiktoken', 
            'nest_asyncio', 
            'beautifulsoup4',
        ],
    },
    author='Ryan Eggleston',
    author_email='kre8mymedia@gmail.com',
    description='A collection of utilities by Prompt Engineers',
    long_description=open('README.md', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/promptengineers-ai/core',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from typing import Union

from promptengineers.config.test import TEST_USER_ID
from promptengineers.config import (OPENAI_API_KEY, ACCESS_KEY_ID, ACCESS_SECRET_KEY, BUCKET,
                            PINECONE_ENV, PINECONE_INDEX, PINECONE_KEY, ALCHEMY_API_KEY,
                            APP_SECRET, MONGO_CONNECTION, REDIS_URL, OLLAMA_BASE_URL, MONGO_DB_NAME)
from promptengineers.utils.security import decrypt

LEGACY_KEY_NAME = 'keys'
NEW_KEY_NAME = 'env'

class UserRepo:
    def __init__(self) -> None:
        pass

    def find(self, user_id: str = None):
        if user_id == TEST_USER_ID:
            return {
                'data': {
                    'env': {
                        'OPENAI_API_KEY': OPENAI_API_KEY,
                        'ACCESS_KEY_ID': ACCESS_KEY_ID,
                        'ACCESS_SECRET_KEY': ACCESS_SECRET_KEY,
                        'BUCKET': BUCKET,
                        'PINECONE_KEY': PINECONE_KEY,
                        'PINECONE_ENV': PINECONE_ENV,
                        'PINECONE_INDEX': PINECONE_INDEX,
                        'ALCHEMY_API_KEY': ALCHEMY_API_KEY,
                        'MONGO_CONNECTION': MONGO_CONNECTION,
                        'MONGO_DB_NAME': MONGO_DB_NAME,
                        'REDIS_URL': REDIS_URL,
                        'OLLAMA_BASE_URL': OLLAMA_BASE_URL,
                    },
                },
            }

    ##################################################
    ## Find API Token
    ##################################################
    def find_token(
        self,
        user_id: str,
        token_names: Union[str, list],
        legacy: bool = False,
    ):
        user = self.find(user_id)

        # Use the first set of conditions if datagrid is None
        if not legacy:
            keys = user.get('data', {}).get(NEW_KEY_NAME, {})

            if isinstance(token_names, str):
                token_value = keys.get(token_names)
                should_decrypt = True if APP_SECRET else False
                return decrypt(token_value) if should_decrypt else token_value
            else:
                return {
                    token_name: (decrypt(keys[token_name]) if APP_SECRET else keys[token_name])
                    for token_name in token_names if token_name in keys
                }

        # Use the second set of conditions if datagrid is provided
        else:
            keys = user.get('data').get(LEGACY_KEY_NAME)
            data = list(filter(lambda x: x != {}, keys))

            if isinstance(token_names, str):
                for item in data:
                    if token_names == item.get('key'):
                        return decrypt(item.get('value'))
            else:
                tokens = {}
                for item in data:
                    key = item.get('key')
                    if key in token_names:
                        tokens[key] = decrypt(item.get('value'))
                return tokens



import nest_asyncio
from langchain.document_loaders import (CSVLoader, DirectoryLoader, GitbookLoader,
										PyPDFLoader, TextLoader,
										UnstructuredHTMLLoader,
										UnstructuredMarkdownLoader,
										UnstructuredURLLoader, WebBaseLoader,
										YoutubeLoader, SitemapLoader, BlockchainDocumentLoader)
from langchain.document_loaders.blockchain import BlockchainType

from promptengineers.strategies.loaders import CopyPasteLoader
from promptengineers.config import ALCHEMY_API_KEY
from promptengineers.utils import get_links, logger

nest_asyncio.apply()

class LoaderFactory:
	@staticmethod
	def create_loader(
		loader_type: ('gitbook', 'web_base', 'yt', 'polygon', 'ethereum', 'sitemap', 'website', 'urls', 'copy', 'txt', 'html', 'md', 'directory', 'csv', 'pdf'),
		body
	):
		if loader_type == 'gitbook':
			urls = body.get('urls', [])
			logger.info('[DocumentLoaderFactory.create_loader] Gitbook Link: %s', urls[0])
			return GitbookLoader(urls[0], load_all_paths=True)

		if loader_type == 'web_base':
			urls = body.get('urls', [])
			logger.info('[DocumentLoaderFactory.create_loader] Web Base: %s', urls[0])
			return WebBaseLoader(urls[0])

		if loader_type == 'yt':
			yt_id = body.get('ytId')
			logger.info('[DocumentLoaderFactory.create_loader] Youtube: https://youtube.com/watch?v=%s', yt_id)
			return YoutubeLoader(yt_id)

		if loader_type == 'polygon':
			contract_address = body.get('contract_address', '')
			logger.info('[DocumentLoaderFactory.create_loader] Polygon Contract Address: %s', contract_address)
			return BlockchainDocumentLoader(
				contract_address=contract_address,
				blockchainType=BlockchainType.POLYGON_MAINNET,
				api_key=ALCHEMY_API_KEY,
			)

		if loader_type == 'ethereum':
			contract_address = body.get('contract_address', '')
			logger.info('[DocumentLoaderFactory.create_loader] Ethereum Contract Address: %s', contract_address)
			return BlockchainDocumentLoader(
				contract_address=contract_address,
				api_key=ALCHEMY_API_KEY,
			)

		if loader_type == 'sitemap':
			urls = body.get('urls', [])
			logger.info('[DocumentLoaderFactory.create_loader] Sitemap: %s', urls)
			return SitemapLoader(web_path=urls[0])

		if loader_type == 'website':
			urls = body.get('urls', [])
			unique_links = get_links(urls[0])
			logger.info('[DocumentLoaderFactory.create_loader] Website: %s', unique_links)
			return UnstructuredURLLoader(urls=unique_links)

		if loader_type == 'urls':
			urls = body.get('urls', [])
			logger.info('[DocumentLoaderFactory.create_loader] URLs: %s', urls)
			return UnstructuredURLLoader(urls=urls)

		if loader_type == 'copy':
			logger.info('[DocumentLoaderFactory.create_loader] Copy %s', body.get("text")[:500] + '...')
			return CopyPasteLoader(text=body.get('text'))

		if loader_type == 'txt':
			logger.info('[DocumentLoaderFactory.create_loader] Text: %s', body.get("file_path"))
			return TextLoader(body.get('file_path'))

		if loader_type == 'html':
			logger.info('[DocumentLoaderFactory.create_loader] HTML: %s', body.get("file_path"))
			return UnstructuredHTMLLoader(body.get('file_path'))

		if loader_type == 'md':
			logger.info('[DocumentLoaderFactory.create_loader] Markdown: %s', body.get("file_path"))
			return UnstructuredMarkdownLoader(body.get('file_path'))

		if loader_type == 'directory':
			logger.info('[DocumentLoaderFactory.create_loader] Directory: %s', body.get("file_path"))
			return DirectoryLoader(body.get('file_path'), glob="**/*")

		if loader_type == 'csv':
			logger.info('[DocumentLoaderFactory.create_loader] CSV: %s', body.get("file_path"))
			return CSVLoader(body.get('file_path'))

		if loader_type == 'pdf':
			logger.info('[DocumentLoaderFactory.create_loader] PDF: %s', body.get("file_path"))
			return PyPDFLoader(body.get('file_path'))

		raise ValueError(f'Unsupported document loader type: {loader_type}')



from langchain.document_loaders.base import BaseLoader
from langchain.docstore.document import Document

class CopyPasteLoader(BaseLoader):
	def __init__(self, text):
		self.text = text

	def load(self):
		return [Document(page_content=self.text)]
